/*********************************************************************

 gssklog.c

Description:

This is a modified version of the OpenAFS klog.c 
See: http://www.openafs.org

We will use the Kerberos GSSAPI or the Microsoft SSPI,
or the Globus Grid Security Interface (GSI) 
to authenticate to a gssklogd and get a Kerberos ticket, 
which will be used as an AFS token.

The program was hightly stripped down to avoid as much
extra AFS routines as possible, as it will be linked with 
the GSI or Kerberos GSSAPI.

When run on windows and using the Microsoft SSPI, a 
user@realm and password may be provided.  With a normal
GSSAPI, credentials must have been obtained using 
kinit or grid-proxy-init for example. 


CVS Information:

    $Source: $
    $Date: $
    $Revision: $
    $Author: $

********************************************************************/

static char *rcsid = "$Header: $";

/* Copyright from OpenAFS: */

/* Copyright (C) 1991, 1989 Transarc Corporation - All rights reserved */
/*
 * (C) COPYRIGHT IBM CORPORATION 1988, 1989
 * LICENSED MATERIALS - PROPERTY OF IBM
 */

#include <afs/param.h>
#include <afs/stds.h>

#ifdef _WIN32
#include <windows.h>
#include <afs/cm_config.h>
#include <winsock2.h>
#include <winsvc.h>
#include "gssklog_wrpw.h"
#define wsclose closesocket
#else
#define wsclose close
#endif

#include <sys/types.h>
#ifdef	AFS_AIX32_ENV
#include <signal.h>
#endif

#include <stdio.h>
#ifndef _WIN32
#include <pwd.h>
#include <netdb.h>
#endif

/* dont even include the com_err for now */
#define com_err(N,C,R) \
  (fprintf(stderr,"%s : %d : %s\n", N, C, R))

#include <afs/dirpath.h>
#include <afs/auth.h>
#include <afs/cellconfig.h>
#include <afs/cmd.h>

#include "gssklog.h"
#include "gssklog_comm.h"

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif

#ifndef DEFAULT_PORT
#define DEFAULT_PORT 750
#endif

#define KABADARGUMENT 2
#define KABADNAME 3

static int doit(AUTH_METHOD * method,
			int more_chances,
			int Silent,
			void * vpcred_handle,
			struct sockaddr_in  * hostAddr,
			char * service_princ_name,
			unsigned char * req,
			int reqlen,
			unsigned char * resp,
			size_t * resplen);

/* gss handles are pointers, sspi are two ulongs, 
 * so we point to the handle instead */

static void * pcred_handle; 
static int gsiklog_comp = 0;

/*
 * We define two methods, the GSSAPI and the SSPI
 * On windows, they both use the GSSAPI protocol,
 * Depending on the machine that this is run on, 
 * it may have one or both, If both we can try both
 * if there are credentials.
 */ 
static int smethod = 0;
static AUTH_METHOD * method[2] = {NULL,NULL};


#ifdef _WIN32
#define KLOGEXIT(code)  if (code) sleep(5); exit(code)
#else
#define KLOGEXIT(code) (exit(code))
#endif

extern int CommandProc (
  struct cmd_syndesc *as,
  char *arock
);

static int zero_argc;
static char **zero_argv;

/* This performs the backslash quoting defined by AC_ParseLoginName. */

static char map_char (str, ip)
  char *str;
  int  *ip;
{   char c = str[*ip];
    if (c == '\\') {
	c = str[++(*ip)];
	if ((c >= '0') && (c <= '7')) {
	    c = c - '0';
	    c = (c * 8) + (str[++(*ip)] - '0');
	    c = (c * 8) + (str[++(*ip)] - '0');
	}}
    return c;
}

/* This routine parses a string that might be entered by a user from the
   terminal.  It defines a syntax to allow a user to specify his identity in
   terms of his name, instance and cell with a single string.  These three
   output strings must be allocated by the caller to their maximum length.  The
   syntax is very simple: the first dot ('.') separates the name from the
   instance and the first atsign ('@') begins the cell name.  A backslash ('\')
   can be used to quote these special characters.  A backslash followed by an
   octal digit (zero through seven) introduces a three digit octal number which
   is interpreted as the ascii value of a single character. */

/* Errors:
     KABADARGUMENT - if no output parameters are specified.
     KABADNAME - if a component of the user name is too long or if a cell was
       specified but the cell parameter was null.
 */

afs_int32 ka_ParseLoginName (
  char *login,
  char  name[MAXKTCNAMELEN],
  char  inst[MAXKTCNAMELEN],
  char  cell[MAXKTCREALMLEN])
{
    int  login_len = strlen (login);
    char rc,c;
    int	 i,j;
#define READNAME 1
#define READINST 2
#define READCELL 3
    int	 reading;

    if (!name) return KABADARGUMENT;
    strcpy (name, "");
    if (inst) strcpy (inst, "");
    if (cell) strcpy (cell, "");
    reading = READNAME;
    i = 0;
    j = 0;
    while (i < login_len) {
	rc = login[i];
	c = map_char (login, &i);
	switch (reading) {
	  case READNAME:
	    if (rc == '@') {
		name[j] = 0;		/* finish name */
		reading = READCELL;	/* but instance is null */
		j = 0;
		break;
	    }
	    if (inst && (rc == '.')) {
		name[j] = 0;		/* finish name */
		reading = READINST;
		j = 0;
		break;
	    }
	    if (j >= MAXKTCNAMELEN-1) return KABADNAME;
	    name[j++] = c;
	    break;
	  case READINST:
	    if (!inst) return KABADNAME;
	    if (rc == '@') {
		inst[j] = 0;		/* finish name */
		reading = READCELL;
		j = 0;
		break;
	    }
	    if (j >= MAXKTCNAMELEN-1) return KABADNAME;
	    inst[j++] = c;
	    break;
	  case READCELL:
	    if (!cell) return KABADNAME;
	    if (j >= MAXKTCREALMLEN-1) return KABADNAME;
	    cell[j++] = c;
	    break;
	}
	i++;
    }
    if (reading == READNAME) name[j] = 0;
    else if (reading == READINST) {
	if (inst) inst[j] = 0;
	else return KABADNAME;
    }
    else if (reading == READCELL) {
	if (cell) cell[j] = 0; 
	else return KABADNAME;
    }

    /* the cell is really an authDomain and therefore is really a realm */
    if (cell) ucstring (cell, cell, MAXKTCREALMLEN);
    return 0;
}

int osi_audit(void)
{
    return 0;
}

#ifdef _WIN32
static long 
my_cm_configProc( void *rockp, struct sockaddr_in *addrp, char *namep)
{
	struct afsconf_cell *cellconfig;
	int i;
       
	cellconfig = (struct afsconf_cell *)rockp;
	i = cellconfig->numServers;

        memcpy(&(cellconfig->hostAddr[i]), addrp, sizeof(struct sockaddr_in));
        strncpy((char *)&(cellconfig->hostName[i]), namep, MAXHOSTCHARS);
	cellconfig->numServers = i + 1;
        return 0;
}	
#endif

static char *getpipepass(void)
{
    static char gpbuf[BUFSIZ];
    /* read a password from stdin, stop on \n or eof */
    register int i, tc;
    memset(gpbuf, 0, sizeof(gpbuf));
    for(i=0; i<(sizeof(gpbuf)-1); i++) {
    tc = fgetc(stdin);
    if (tc == '\n' || tc == EOF) break;
    gpbuf[i] = tc;
    }
    return gpbuf;
}

/*******************************************************************/
int main (
  int   argc,
  char *argv[])
{
    struct cmd_syndesc *ts;
    afs_int32 code;
#ifdef _WIN32
	WSADATA WSAjunk;
#endif

#ifdef	AFS_AIX32_ENV
    /*
     * The following signal action for AIX is necessary so that in case of a 
     * crash (i.e. core is generated) we can include the user's data section 
     * in the core dump. Unfortunately, by default, only a partial core is
     * generated which, in many cases, isn't too useful.
     */
    struct sigaction nsa;
    
    sigemptyset(&nsa.sa_mask);
    nsa.sa_handler = SIG_DFL;
    nsa.sa_flags = SA_FULLDUMP;
    sigaction(SIGABRT, &nsa, NULL);
    sigaction(SIGSEGV, &nsa, NULL);
#endif
    zero_argc = argc;
    zero_argv = argv;

#ifdef _WIN32
	WSAStartup(0x0101, &WSAjunk);
#endif

    ts = cmd_CreateSyntax((char *) 0, CommandProc, 0, "obtain tokens using GSS authentication");

#define aXFLAG 0
#define aPRINCIPAL 1

#ifdef _WIN32
#define aPASSWORD 2
#define aOFFSET 2
#else
#define aOFFSET 1  
#endif

#define aCELL 1 + aOFFSET
#define aPIPE 2 + aOFFSET
#define aSILENT 3 + aOFFSET
#define aLIFETIME 4 + aOFFSET
#define aSETPAG 5 + aOFFSET
#define aPORT 6 + aOFFSET
#define aGSICMP 7 + aOFFSET
#define aSERVER 8 + aOFFSET
#define aPATH 9 + aOFFSET
/* next may be ifdefed */
#define aMSSSPI 10 + aOFFSET


    cmd_AddParm(ts, "-x", CMD_FLAG, CMD_OPTIONAL, "(obsolete, noop)");
    cmd_Seek(ts, aPRINCIPAL);
    cmd_AddParm(ts, "-principal", CMD_SINGLE, CMD_OPTIONAL, 
			"user or user@realm");
#ifdef _WIN32
	cmd_AddParm(ts, "-password", CMD_SINGLE, CMD_OPTIONAL, 
			"user's password");
#endif
    cmd_AddParm(ts, "-cell", CMD_SINGLE, CMD_OPTIONAL, "cell name");
	cmd_AddParm(ts, "-pipe", CMD_FLAG, CMD_OPTIONAL, 
			"read password from stdin");
    cmd_AddParm(ts, "-silent", CMD_FLAG, CMD_OPTIONAL, "silent operation");
    cmd_AddParm(ts, "-lifetime", CMD_SINGLE, CMD_OPTIONAL, "ticket lifetime in hh[:mm[:ss]]");
    cmd_AddParm(ts, "-setpag", CMD_FLAG, CMD_OPTIONAL, "Create a new setpag before authenticating");

	cmd_AddParm(ts, "-port", CMD_SINGLE, CMD_OPTIONAL, "port");

	cmd_AddParm(ts, "-gsiklog", CMD_FLAG, CMD_OPTIONAL, "Use old gsiklog ");
	cmd_AddParm(ts, "-server", CMD_SINGLE, CMD_OPTIONAL, "server");

	cmd_AddParm(ts, "-p", CMD_SINGLE, CMD_OPTIONAL, 
		"Get token for cell containing this path");

#ifdef _WIN32
	cmd_AddParm(ts, "-ms", CMD_FLAG, CMD_OPTIONAL,"Use MS credentials via sspi before MIT Kerberos");
#endif

	/* 
	 * setup the methods to try. 
	 * method[0] with be MIT GSSAPI if available. 
	 * method[1] will be MS SSPI if available.
	 */

	 method[0] = gssklog_gss_init_method();
#ifdef _WIN32
	 method[1] = gssklog_sspi_init_method();
#endif

#ifdef DEBUG
		fprintf(stderr,"methods found: %p %p\n",method[0], method[1]);
#endif

	if (method[0] == NULL && method[1] == NULL) {
		fprintf(stderr," No GSS or SSPI methods available\n");
		KLOGEXIT( KABADARGUMENT );
	} 

    code = cmd_Dispatch(argc, argv);
    KLOGEXIT(code);
}

/*********************************************************
Function:  doit()

Description:
    Contact the host, authenticate, send the request,
    and receive the token.  If we fail, we may be called
    again to try a replicated server.

Parameters:

Return codes:
    0    worked, we have a returned response
    < 0  failed, and don't try another replicate server.
    > 0  failed, try the next server.
***********************************************************************/
static int 
doit(AUTH_METHOD * method,
		int more_chances,
		int Silent,
		void * vpcred_handle,
		struct sockaddr_in  * hostAddr,
		char * service_princ_name,
		unsigned char * req,
		int reqlen,
		unsigned char * resp,
		size_t * resplen)
{

    const char this_function[] = "doit";
	OM_uint32 major_status;
	OM_uint32 minor_status;
	void * pcontext_handle = NULL;
	gss_buffer_desc   input_buffer_desc = {0,NULL};
	gss_buffer_desc   output_buffer_desc = {0,NULL};
	gss_buffer_t      output_buffer = &output_buffer_desc ;
	gss_buffer_t      input_buffer  = &input_buffer_desc ;

	int peer_socket = -1;
	int rc;
	int msg_result;

/* 
 * At this point we now know where the server is,
 * and if it is replicated. 
 * For now we expect service, host, port to be parameters. 
 */

	peer_socket = connect_to_server_sockaddr(hostAddr);
	if (peer_socket < 0) {
		if (!more_chances) {
			perror("Connecting to server"); 
		}
		rc = 1;
		goto err;
	}

#ifdef DEBUG
  fprintf(stderr, "N %s: Connected to acceptor\n", this_function) ;
#endif 

/*DEE call the auth_context_establishment */

	major_status = method->auth_init_sec_context(method,
		&minor_status,
		vpcred_handle,
		&pcontext_handle,
		peer_socket,
		service_princ_name);


	if ( major_status != 0) {
		if (!Silent) {
			method->auth_display_status(method,
				"init_sec_context failed",
				major_status, minor_status);
		}
		if (method->version == 1 &&  minor_status == -1)
			rc = -1;
		else  
			rc = 2;
		goto err;
	}

    output_buffer->length = reqlen;
    output_buffer->value = req;

/* send the request */

    msg_result = send_message(method,
                  peer_socket,
                  &pcontext_handle,
                  output_buffer,
                  1, /* 1 =  wrap */
                  1) ; /* 1 = encrypt */

    /* since we are sending a constant string, we dont release the buffer */
    /* gss_release_buffer(&minor_status, output_buffer); */

    if ( msg_result < 0 ) {
      rc = 6;
      goto err;
    }
#ifdef DEBUG
    fprintf(stderr,
        "N %s: Data sent successfully.\n",
        this_function) ;
#endif

    /* read in the response */
      msg_result = receive_message(method,
                   peer_socket,
                   &pcontext_handle,
                   input_buffer,
                   1) ; /* 1 = unwrap */
      if ( msg_result < 0 ) {
        rc = 7;
        goto err;
      }

      if (!input_buffer->value || input_buffer->length > *resplen) {
        rc = 8;
        goto err;
      }

      *resplen = input_buffer->length;
      memcpy(resp, input_buffer->value, *resplen);

    rc = 0;

err:
	
    if (pcontext_handle)
    method->auth_delete_sec_context(method,
			&minor_status,
            &pcontext_handle);

    if (peer_socket >= 0)
        wsclose(peer_socket) ;
    return rc;

}

/*************************************************************/

int CommandProc (
  struct cmd_syndesc *as,
  char *arock)
{
    char  name[MAXKTCNAMELEN];
    char  instance[MAXKTCNAMELEN];
    char  pcell[MAXKTCREALMLEN];
    char  cell[MAXKTCREALMLEN];
	char * username = NULL;
	char * password = NULL;
	char passwd[BUFSIZ];
#ifdef _WIN32
	char  defaultCell[256];
#endif
	char  service_princ_name[MAXHOSTNAMELEN+9];
	char  server[MAXHOSTNAMELEN];
	struct sockaddr_in serverAddr;
	struct hostent *hp ;

#ifndef _WIN32
	struct afsconf_dir *configdir;
#endif
	struct afsconf_cell cellconfig;


	int foundExplicitCell = 0;
	int foundExplicitServer = 0; 
    int	  code=0;
	OM_uint32 major_status[2] = {-1,-1};
	OM_uint32 minor_status[2] = {0,0};
	int try_method[2] = {-1,-1};
    int   i, dosetpag;
    afs_int32  lifetime = 0;			/* requested ticket lifetime */
	unsigned char request[BUFSIZ];
	int  requestlen;
	unsigned char answer[BUFSIZ];
	size_t answerlen;

    static char	rn[] = "gssklog";		/*Routine name*/
	static int Silent = 0;      /* Don't want error messages */

    char *reason;			/* string describing errors */
	short port = DEFAULT_PORT;/* using K4 TCP port for now */

    /* blow away command line arguments */
    for (i=1; i<zero_argc; i++) memset(zero_argv[i], 0, strlen(zero_argv[i]));
    zero_argc = 0;

/*
 * Make sure the service is running, at least try and start if not
 */
#ifdef DOES_NOT_WORK_WIN32
	{
		SC_HANDLE hManager;
		SC_HANDLE hService;
		SERVICE_STATUS Status;

		if ((hManager = OpenSCManager (NULL, NULL,
					SC_MANAGER_CONNECT)) != NULL) {
			if ((hService = OpenService (hManager, 
						TEXT("TransarcAFSDaemon"),
						SERVICE_START|SERVICE_QUERY_STATUS)) != NULL) {
				if (QueryServiceStatus (hService, &Status)) {
					if (Status.dwCurrentState == SERVICE_STOPPED) { 
						if (StartService (hService, 0, 0)) {
							Sleep(10000); /* wait for 10 sec */
						}
					}
					CloseServiceHandle (hService);
				}
				CloseServiceHandle (hManager);
			}
		}
	}
#endif

#ifndef _WIN32
	if (!(configdir = afsconf_Open(AFSDIR_CLIENT_ETC_DIRPATH))) {
		fprintf(stderr,"Unable to open AFS config file\n");
		KLOGEXIT( KABADARGUMENT );
	}
#endif

    /* first determine quiet flag based on -silent switch */
    Silent = (as->parms[aSILENT].items ? 1 : 0);

    /* Determine if we should also do a setpag based on -setpag switch */
    dosetpag = (as->parms[aSETPAG].items ? 1 : 0);

    strcpy (instance, "");

    /* Parse our arguments. */

	if (as->parms[aPATH].items) {
		/* ignore for now  and use cell of host */
	}

    if (as->parms[aCELL].items) {
		strncpy (cell, as->parms[aCELL].items->data, sizeof(cell));
		foundExplicitCell = 1;
    } else {
#ifdef _WIN32
		cm_GetRootCellName(defaultCell);
        strncpy(cell, defaultCell, sizeof(cell));
#else
		if (afsconf_GetLocalCell(configdir, cell, sizeof(cell))) {
			fprintf(stderr,"Cant get local cell name\n");
			KLOGEXIT( KABADARGUMENT );
		}
#endif
	}

#ifdef _WIN32
	if (as->parms[aPASSWORD].items) {
		password = as->parms[aPASSWORD].items->data;
	}
#endif

    if (as->parms[aPRINCIPAL].items) {
		/* we will pass in full principal name */
		ka_ParseLoginName (as->parms[aPRINCIPAL].items->data,
				   name, instance, pcell);
		if (strlen (instance) > 0) {
	   		 if (!Silent) {
			fprintf (stderr, 
			 "Non-null instance (%s) may cause strange behavior.\n",
			 instance);
	   	   }
		}
#ifdef _WIN32
		username = as->parms[aPRINCIPAL].items->data;
		if (!password) {
			if (as->parms[aPIPE].items) {
				strncpy(passwd, getpipepass(), sizeof(passwd));
				password = passwd;
			} else {
				char prompt[BUFSIZ];
				snprintf(prompt,sizeof(prompt),
					"Enter password for %s",
					username);
				save_instance(0); /*DEE may need work */
			
				if (read_passphrase_win32(prompt, 
						passwd, sizeof(passwd)) == 0) {
					fprintf (stderr, 
						"Unable to get password\n");
					KLOGEXIT( KABADARGUMENT );
				}
				password = passwd;
			}
		}
#endif
    } else {
		/* No explicit name provided: use gssapi names */
		/* should use the "princ_from_cert" */
		/* lets gssklogd do a mapping based on gss-cleint-name */
		strncpy(name,"princ_from_cert",sizeof(name));
    }


    if (as->parms[aLIFETIME].items) {
	char *life = as->parms[aLIFETIME].items->data;
	char *sp;			/* string ptr to rest of life */
	lifetime = 3600*strtol (life, &sp, 0); /* hours */
	if (sp == life) {
bad_lifetime:
	    if (!Silent) fprintf (stderr, "%s: translating '%s' to lifetime failed\n",
			       rn, life);
	    return KABADARGUMENT;
	}
	if (*sp == ':') {
	    life = sp+1;		/* skip the colon */
	    lifetime += 60*strtol (life, &sp, 0); /* minutes */
	    if (sp == life) goto bad_lifetime;
	    if (*sp == ':') {
		life = sp+1;
		lifetime += strtol (life, &sp, 0); /* seconds */
		if (sp == life) goto bad_lifetime;
		if (*sp) goto bad_lifetime;
	    } else if (*sp) goto bad_lifetime;
	} else if (*sp) goto bad_lifetime;
	if (lifetime > MAXKTCTICKETLIFETIME) {
	    if (!Silent) fprintf (stderr, "%s: a lifetime of %.2f hours is too long, must be less than %d.\n", rn, (double)lifetime/3600.0, MAXKTCTICKETLIFETIME/3600);
	    KLOGEXIT( KABADARGUMENT );
	}
    } else lifetime = 0;

	/* get port  for debuging, or if more then one gssklogd is needed */

	if (as->parms[aPORT].items) {
		port =   atoi(as->parms[aPORT].items->data);
	}

	/* Use explicit server, Mostly for testing, or if server is not on afsdb machine */
	if (as->parms[aSERVER].items) {
		strncpy (server, as->parms[aSERVER].items->data, sizeof(server));
		foundExplicitServer = 1;
	}

	/* Make this make this compatable with older gsiklog */

	gsiklog_comp = (as->parms[aGSICMP].items ? 1 : 0);

	smethod = 0;
#ifdef _WIN32
	smethod = (as->parms[aMSSSPI].items ? 1 : 0);
#endif
	
/* DEE needs work on if method available, and which to try first */

	/* Now have all the parameters */

#ifdef _WIN32
	memset(&cellconfig, 0, sizeof(cellconfig));
	strncpy(cellconfig.name, cell, sizeof(cellconfig.name));

	if (cm_SearchCellFile(cellconfig.name, NULL,
			my_cm_configProc, &cellconfig)) 
#else
	if (afsconf_GetCellInfo(configdir, cell, NULL, &cellconfig)) 
#endif
	{
		fprintf(stderr,"Unable to get Cell Info\n");
		KLOGEXIT( KABADARGUMENT );
	}
#ifdef DEBUG
	fprintf(stderr,"found cell=%s\n",cell);
#endif
	
	/* 
	 * We will try the method the user requested first, if it exists,
	 * if it does not have creds, we will try the other.
	 * smethod is 0 for gss, and 1 for sspi. the -ms option set it to 1
	 * We know we have at least 1 method, we checked earlier.
	 * If it fails, save the codes till after we try the other,
	 * then report the errors.
	 */

	if (!method[smethod]) smethod = smethod? 0 : 1; /* method not loaded*/
	try_method[0] = smethod;
	major_status[0] = method[smethod]->auth_acquire_cred(method[smethod], &minor_status[0], &pcred_handle, username, password);
#ifdef _WIN32
	if (major_status[0]) {
		smethod = smethod? 0 : 1;
		if  (method[smethod]) { 
			try_method[1] = smethod;
			major_status[1] = method[smethod]->auth_acquire_cred(method[smethod], &minor_status[1], &pcred_handle, username, password);
		}
	}
#endif

#ifdef DEBUG
	fprintf (stderr,"smethod=%d try-maj-min (%d %d %d) (%d %d %d)\n",
		smethod,try_method[0],major_status[0],minor_status[0],
		        try_method[1],major_status[1],minor_status[1]);
#endif
	if (major_status[0] && major_status[1]) {
		if (!Silent) {
			if (try_method[0] >= 0) {
				method[try_method[0]]->auth_display_status(method[try_method[0]], "Unable to get user credentials", major_status[0], minor_status[0]);
			}
			if (try_method[1] >= 0) {
				method[try_method[1]]->auth_display_status(method[try_method[1]], "Unable to get user credentials", major_status[1], minor_status[1]);
			}
		}
		KLOGEXIT( KABADARGUMENT );
	}
/* smethod has the method that worked. */

/* We may want to get two tokens, one for the local cell */
/* and one for the cell that contains the path */
/* TODO: DEE loop here over the cells. Change KLOGEXIT */
/* below to break out of loop. */

/* make up the simple request */

	requestlen = sizeof(request);
	if (code = create_request(request, &requestlen, 
				name, instance, cell, 
				"afs", "", cell, lifetime)) {
		if (!Silent) {
			fprintf(stderr,"Unable to make a request buffer\n" );
		}
		KLOGEXIT( KABADARGUMENT );
	}

/* Try explictserver */

	if ( foundExplicitServer ) {
		if (gsiklog_comp) {
			/* use afs@cell */
			strcpy(service_princ_name,"afs@");
			strcat(service_princ_name,cell);
		} else {
			strcpy(service_princ_name,"gssklog@");
			strcat(service_princ_name,server);
		}
		/* Look up addr of server */

		if ((hp = gethostbyname(server)) == NULL) {
			fprintf(stderr, "Unknown host: %s\n", server);
			code = -1;
		} else {
			serverAddr.sin_family = hp->h_addrtype;
			memcpy((char *)&serverAddr.sin_addr, hp->h_addr, hp->h_length);
			serverAddr.sin_port = htons(port);
			answerlen = sizeof(answer);
			code = doit(method[smethod],
						0,
						Silent,
						pcred_handle,
						&serverAddr,
						service_princ_name,
						request, requestlen,
						answer, &answerlen);

		}
	} else /* Assume the server is running on one or more of the AFSDB machines */
	
	for (i = cellconfig.numServers-1; i>=0; i--) {
		if (gsiklog_comp) {
			/* use afs@cell */
			strcpy(service_princ_name,"afs@");
			strcat(service_princ_name,cell);
		} else {
			strcpy(service_princ_name,"gssklog@");
                        strcat(service_princ_name,cellconfig.hostName[i]);
		}

		answerlen = sizeof(answer);
		cellconfig.hostAddr[i].sin_port = htons(port);
		code = doit(method[smethod],
				i, 
				Silent,
				pcred_handle,
				&cellconfig.hostAddr[i],  
				service_princ_name,
				request, requestlen, 
				answer, &answerlen);

		if (code <= 0) {
			break;
		}
		if (!Silent) {
			fprintf(stderr,"Problem %d with server %s%s\n",
				code,
				cellconfig.hostName[i],
				i? ", trying next" : "");
		}
	}

	/* did we get a token */
	if (code) {
		if (!Silent) {
			fprintf(stderr,"Failed code = %d\n",code);
		}
		KLOGEXIT( KABADARGUMENT );
	}

	/* need to stuff into kernel */
	
	if (code = process_response(answer, answerlen,
			cell,
			&reason,  dosetpag)) {
		if (!Silent) {
	   	 fprintf (stderr,
		     "Unable to get token: %s.\n", 
				reason?reason:"unknown problems");
	  	}
		KLOGEXIT( code );
      }

 
    return 0;
}
