/*****************************************************************

 gssklog_comm.c : 


Description:
	Common routines used with the gssklog. IO and error.

CVS Information:

	$Source: $
	$Date: $
	$Revision: $
	$Author: $

*******************************************************************/

static char *rcsid = "$Header: $";

/**********************************************************************
                             Include header files
**********************************************************************/
#ifdef _WIN32
#include <windows.h>
#include <winsock2.h>
#define wsread(a,b,c) recv(a,b,c,0)
#define wswrite(a,b,c) send(a,b,c,0)
#define wsclose closesocket
#else
#define wsread read
#define wswrite write
#define wsclose close
#endif

#include <stdio.h>
#include <sys/types.h>

#ifndef _WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <fcntl.h>
#endif

#include <errno.h>
#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "gssklog_comm.h"

/**********************************************************************
                               Type definitions
**********************************************************************/

/**********************************************************************
                          Module specific prototypes
**********************************************************************/

int connection(char *, unsigned int, char *);
int connect_to_server(char *, unsigned int);
int create_socket(unsigned int);
int setup_connection();
int establish_connection( int);
int read_all(int fildes, void *buf, size_t nbyte);

/**********************************************************************
                       Define module specific variables
**********************************************************************/
/********************************************************************
 Function: read_all
********************************************************************/
int read_all(int fildes, void *buf, size_t nbyte)
{
     size_t ret = 0;
     char *ptr;
#ifndef WIN32
	 fd_set fdsets;
	 struct timeval waitlen;
#endif

     for (ptr = (char *) buf; nbyte; ptr += ret, nbyte -= ret) {
	  ret = wsread(fildes, (void *) ptr, nbyte);
	  if (ret == -1) {
		if (errno == EINTR) {
			ret = 0;
		   	continue;
		}
#ifndef WIN32
		else if (errno == EAGAIN) {
			waitlen.tv_usec = 0;
			waitlen.tv_sec = 60; /* Mitigate DOS attack, 60 max - GSI may take more the 10 sec */
			FD_ZERO(&fdsets);
			FD_SET(fildes, &fdsets);
			ret = select(fildes+1, &fdsets, 0, 0, &waitlen);
			if (ret == 0) { /* timeout */
				return -2;
			}
			if (ret == -1) { 
				if (errno == EINTR) {
					ret = 0;
					continue;
				} 
				else if (errno == EBADF) {
					return -3; /* file may be closed */
				}
				perror("select failed");
				return -4;
			}
			ret = 0;
			continue;
		}
#endif
		perror("read_all failed");
		return -5;
	  } else if (ret == 0) { /* socket must have closed */
	       return(ptr-(char *) buf);
	  }
     }

     return(ptr-(char *)buf);
}

/*********************************************************************
 * Function: connect_to_server (client )
 *
 * Purpose: Opens a TCP connection to the name host and port.
 *
 * Arguments:
 *
 * 	host		(r) the target host name
 * 	port		(r) the target port, in host byte order
 *
 * Returns: the established socket file desciptor, or -1 on failure
 *
 * Effects:
 *
 * The host name is resolved with gethostbyname(), and the socket is
 * opened and connected.  If an error occurs, an error message is
 * displayed and -1 is returned.
 ******************************************************************/
int connect_to_server(char *host, unsigned int port)
{
  char this_function[] = "connect_to_server";
  struct sockaddr_in saddr;
  struct hostent *hp;
  short sport = port;
#ifdef DEBUG
  fprintf(stderr, "N %s: args\n\t host = %s, port = %d\n", 
	  this_function, host, port);
#endif /* DEBUG */     
  if ((hp = gethostbyname(host)) == NULL) {
    fprintf(stderr, "Unknown host: %s\n", host);
    return -1;
  }
     
  saddr.sin_family = hp->h_addrtype;
  memcpy((char *)&saddr.sin_addr, hp->h_addr, hp->h_length);
  saddr.sin_port = htons(sport);

  return connect_to_server_sockaddr((struct sockaddr_in *) &saddr);
}

/*****************************************************************/
int
connect_to_server_sockaddr(struct sockaddr_in * saddr)
{
  char this_function[] = "connect_to_server_sockaddr";
 
  int s;
  int retval;
	
  if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    perror("creating socket");
    return -1;
  }
#ifdef DEBUG
  fprintf(stderr, "N %s attempting connection to ",
	  this_function);
  {
	int i;
	unsigned char * cp;
	cp = (char *)&(saddr->sin_addr);
	for(i=0;i<4;i++) {
		fprintf(stderr,"%d.", *cp++);
	}
	fprintf(stderr,"\n");
  }
#endif
  retval=connect(s, (struct sockaddr *)saddr, 
				sizeof(struct sockaddr));
  if (retval < 0) {
    return retval;
  }

#ifndef _WIN32
 /*DEE */ fcntl(s, F_SETFL, O_NONBLOCK);
#endif
#ifdef DEBUG
  fprintf(stderr, "N %s connected socket\n",
	  this_function);
#endif /* DEBUG */ 
  return s;
} /* connect_to_server() */

/* client function */
int connection(char *host,unsigned int port, char *service_name)
{
  const char this_function[] = "connection()" ;
  int other_socket;
  if ((other_socket = connect_to_server(host, port)) < 0) {
    fprintf(stderr,"E %s: Couldn't connect to socket\n", this_function);
    return -1;
  }
  return other_socket;

} /* connection() */


/* server function */

/*********************************************************************
 * Function: create_socket
 *
 * Purpose: Opens a listening TCP socket.
 *
 * Arguments:
 *
 * 	port		(r) the port number on which to listen
 *
 * Returns: the listening socket file descriptor, or -1 on failure
 *
 * Effects:
 *
 * A listening socket on the specified port and created and returned.
 * On error, an error message is displayed and -1 is returned.
 *******************************************************************/
int create_socket(unsigned int port)
{
     struct sockaddr_in saddr;
     int s;
	 short sport = port;

     memset(&saddr,0,sizeof(saddr)); /* purify: umr */

     saddr.sin_family = AF_INET;
     saddr.sin_port = htons(sport);
     saddr.sin_addr.s_addr = INADDR_ANY;

     if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
	  perror("creating socket");
	  return -1;
     }
     {
	 int opt;
	 opt = 1;
	 setsockopt (s,SOL_SOCKET,SO_REUSEADDR,(char *)&opt,sizeof(opt));
     }

     if (bind(s, (struct sockaddr *) &saddr, sizeof(saddr)) < 0) {
	  perror("binding socket");
	  return -1;
     }
     if (listen(s, 5) < 0) {
	  perror("listening on socket");
	  return -1;
     }
     return s;
} /* create_socket() */

/******************************************************************/
/* server functonn */
int setup_connection(int port) {
    int my_socket;

    if ((my_socket = create_socket(port)) < 0)
	exit(1);
    return my_socket;
} /* setup_connection() */

/******************************************************************/
/* server function */
int establish_connection(int my_socket) {
  char this_function[] ="establish_connection";
  int	           fromlen;
  int                skt2;
  int                gotit;
  struct sockaddr_in from;

  fromlen = sizeof(from);
  gotit = 0;
  while (!gotit)
  {
    skt2 = accept(my_socket, (struct sockaddr *) &from, &fromlen);
    if (skt2 == -1)
    {
        if (errno == EINTR)
        continue;
    	else
		{
      		fprintf(stderr,
	      		"E %s: Accept failed errno: %d\n", this_function, errno) ;
      		return -1;
    	}
	}
    else
        gotit = 1;
  }

  return(skt2);
} /* establish_connection() */


/********************************************************************/
int 
send_message(AUTH_METHOD * method,
		 int peer_socket,
	     void * pcontext_handle,
	     gss_buffer_t input_message_buffer,
	     int wrap_flag,
	     int encrypt_flag)
{

  const char this_function[] = "send_message()" ;
  gss_buffer_desc output_message_buffer_desc ;
  gss_buffer_t output_message_buffer = &output_message_buffer_desc ;

  OM_uint32 major_status ;
  OM_uint32 minor_status ;

#ifdef DEBUG
    fprintf(stderr, 
	    "N %s: Sending %s %s data: [%d] \n", 
	    this_function, 
		wrap_flag?"wrapped":"",
		encrypt_flag?"encrypted":"",
	    input_message_buffer->length) ;
#endif

  if (wrap_flag) {
    int conf_state ; 
    	major_status = method->auth_wrap(method,
			&minor_status,
	   	 	pcontext_handle,
	   	  	encrypt_flag,
	     	GSS_C_QOP_DEFAULT,
	     	input_message_buffer,
	     	&conf_state,
	     	output_message_buffer) ;
			
	if (major_status != GSS_S_COMPLETE) {
		method->auth_display_status(method, "send_message wrap:",
			major_status, minor_status);
		return -1;
	}

#ifdef DEBUG
    fprintf(stderr, 
	    "N %s: Wrapped data: [%d] \n", 
		this_function,
		output_message_buffer->length);
#endif  

  } else {
    output_message_buffer->length = input_message_buffer->length ;
    output_message_buffer->value = 
      (void*) malloc(output_message_buffer->length) ;

    if (output_message_buffer->value == NULL) {
      fprintf(stderr,
	      "E %s: Could not allocate enough memory! %d\n", 
			this_function, output_message_buffer->length) ;
      return -1; 
    }
    memcpy(output_message_buffer->value, 
	   input_message_buffer->value, 
	   output_message_buffer->length) ;
  }

  {
	unsigned char net_buf_size[4];
    size_t num_written;

    net_buf_size[0] =  output_message_buffer->length >> 24;
    net_buf_size[1] =  output_message_buffer->length >> 16;
    net_buf_size[2] =  output_message_buffer->length >>  8;
    net_buf_size[3] =  output_message_buffer->length;

    num_written = wswrite(peer_socket,
			(void *) &net_buf_size, 
			sizeof(net_buf_size));
    if (num_written != sizeof(net_buf_size) ) {
      fprintf(stderr, 
	      "E %s: Buffersize send_to_server. %d bytes written\n",
	      this_function, num_written) ;
      return -1 ;
    }

    num_written = wswrite(peer_socket,
			output_message_buffer->value,
			output_message_buffer->length);

    if (num_written != output_message_buffer->length) {
      fprintf(stderr, 
	      "E %s: Message not sent correctly.  Only %d bytes written.\n", 
	      this_function, num_written) ;
      return -1 ;
    }

#ifdef DEBUG
    fprintf(stderr, 
	    "N %s: Message sent [%d].\n", 
	    this_function, num_written) ;
#endif
    method->auth_release_buffer(method, &minor_status, output_message_buffer) ;
    return num_written ;
  }
}

/******************************************************************/
int 
receive_message(AUTH_METHOD * method,
		 int peer_socket,
	     void * pcontext_handle,
	     gss_buffer_t output_message_buffer,
		 int wrap_flag)
{
  const char this_function[] = "receive_message()" ;  
  OM_uint32 major_status ;  
  OM_uint32 minor_status ;  
  gss_buffer_desc input_message_buffer_desc ;
  gss_buffer_t input_message_buffer = &input_message_buffer_desc; 

  {
	unsigned char net_buf_size[4];
    size_t num_read ;
    int bufsize;
        
    num_read= read_all(peer_socket, (void *) &net_buf_size, sizeof(net_buf_size));
	if (num_read == -2) {
		return -2; /* looks like a timeout */
	}
	if (num_read == -3) {
		return -3; /* looks like closed connection */
	}
    if (num_read != sizeof(net_buf_size)) {
      fprintf(stderr, 
	      "E %s: Incorrect buf_size read: [%d]\n",
	      this_function, 
	      num_read);
      return -1 ;
    }
  bufsize = (  ( ((unsigned int) net_buf_size[0]) << 24)
             | ( ((unsigned int) net_buf_size[1]) << 16)
             | ( ((unsigned int) net_buf_size[2]) << 8)
             |   ((unsigned int) net_buf_size[3]) );

	/* do sanity check, may be a scan program testing us */
	if (bufsize < 1 || bufsize > 2000000) {
		fprintf(stderr,
		"E %s: Invalid message size %d %2.2x%2.2x%2.2x%2.2x\n",
			this_function, bufsize, 
					net_buf_size[0],
					net_buf_size[1],
					net_buf_size[2],
					net_buf_size[3]);
		return -4; 
	}

	input_message_buffer->length = bufsize;
    input_message_buffer->value = (void*) malloc(input_message_buffer->length);
    if (input_message_buffer->value == NULL) {
      fprintf(stderr,
	      "E %s: Could not allocate enough memory!  %d %8.8x \n", 
			this_function, num_read, bufsize) ;
      input_message_buffer->length = 0;
      return -4;
    }        
    num_read= read_all(peer_socket, 
		   (void *) input_message_buffer->value, 
                   input_message_buffer->length);

	if (num_read == -2) {
		return -2; /* looks like a timeout */
	}
	if (num_read != input_message_buffer->length) {
		fprintf(stderr,
			"E %s: Could not read full token\n", this_function);
			return -1;
	}
    /* unwrap the mesage */
    if (wrap_flag) {

      int conf_state ;
      gss_qop_t qop_state ;


#ifdef DEBUG
      fprintf(stderr, 
	      "N %s: Received a wrapped packet: [%d]\n",
	      this_function, 
	      input_message_buffer->length) ;
#endif

 		  major_status = method->auth_unwrap(method,
			 &minor_status,
			 pcontext_handle,
			 input_message_buffer,
			 output_message_buffer,
			 &conf_state,
			 &qop_state) ;

	  if (major_status != GSS_S_COMPLETE) {
		  method->auth_display_status(method,
			"send_message wrap:",
			major_status, minor_status);
		return -1;
	  }

#ifdef DEBUG
      fprintf(stderr, 
	      "N %s: Message unwrapped.  It was determined to be: %s\n",
	      this_function, 
	      conf_state?"Encrypted":"Clear text") ;
#endif
    } else {
      output_message_buffer->length = input_message_buffer->length ;
      output_message_buffer->value = 
	(void*) malloc(output_message_buffer->length ) ;
      if (output_message_buffer->value == NULL) {
	fprintf(stderr,
		"E %s: Could not allocate enough memory!\n", this_function) ;
	return -1 ; 
      }
      memcpy(output_message_buffer->value,
	     input_message_buffer->value,
	     output_message_buffer->length) ;
    }
#ifdef DEBUG
    fprintf(stderr, 
            "N %s: Received message: [%d]\n",
            this_function, 
            output_message_buffer->length) ;
#endif

  /*   gssklog_release_buffer(smethod,&minor_status, input_message_buffer) ;
   *	let call do it */

    return num_read ;
  }

} /* receive_message */

/*****************************************************************
Function: display_status

Description:
	display the gss errors on stderr. 
******************************************************************/

