/*
 * Copyright 1994 by OpenVision Technologies, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appears in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that the name of OpenVision not be used
 * in advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission. OpenVision makes no
 * representations about the suitability of this software for any
 * purpose.  It is provided "as is" without express or implied warranty.
 *
 * OPENVISION DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL OPENVISION BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
 * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef UNICODE
#undef UNICODE
#endif
#include <windows.h>
#include <rpc.h>
#include <winsock2.h>
#include <windns.h>
#define wsclose closesocket
#define SECURITY_WIN32
#include <Security.h>
#include <NTsecAPI.h>
#include <Sspi.h>

#include "gssklog_comm.h"

#include <stdio.h>
#include <sys/types.h>
#include <malloc.h>
	
/********************************************************************/

static OM_uint32
gssklog_sspi_acquire_cred(AUTH_METHOD * method,
		OM_uint32 *minor_status,
		void ** vpcred_handle,
		char * username,
		char * password)
{
    OM_uint32 maj_stat;
    TimeStamp expiry;
    PCredHandle  pcred_handle = NULL;
    CredHandle cred_handle = {0,0};
    SEC_WINNT_AUTH_IDENTITY_EXA AuthIdentity; /* no cleanup */
    char * dom = NULL;
 
    *vpcred_handle = NULL;
    *minor_status = 0;
 
	/*
	 * SSPI will allow us to get a new cred based on username@dom
	 * and password. 
	 * Otherwise use the login credentials. 
	 */  

	if (username) {
		if (!password) {  /* if username, must have password too */
			return SEC_E_NO_CREDENTIALS;
		}

		memset(&AuthIdentity, 0, sizeof(AuthIdentity));
	
		AuthIdentity.Version = SEC_WINNT_AUTH_IDENTITY_VERSION;
		AuthIdentity.Length = sizeof(AuthIdentity);
		AuthIdentity.Flags = SEC_WINNT_AUTH_IDENTITY_ANSI;
		
        AuthIdentity.User = username;
        dom = strchr(username,'@');
        if (dom) {
            AuthIdentity.UserLength = dom - username;
            dom++;
        } else {
            AuthIdentity.UserLength = strlen(AuthIdentity.User);
        }
    	if (dom) {
        	AuthIdentity.Domain = dom;
        	AuthIdentity.DomainLength = strlen(AuthIdentity.Domain);

    	}
		AuthIdentity.Password = password;
		AuthIdentity.PasswordLength = strlen(AuthIdentity.Password);
	} 

    maj_stat = AcquireCredentialsHandle(
                  username?username:NULL,     // no principal name
                  MICROSOFT_KERBEROS_NAME_A,  // package name
                  SECPKG_CRED_OUTBOUND,
                  NULL,                       // no logon id
                  username?&AuthIdentity:NULL,// auth data
                  NULL,                       // no get key fn
                  NULL,                       // noget key arg
                  &cred_handle,
                  &expiry
                 );

    if (maj_stat == SEC_E_OK)
    {
		pcred_handle = (CredHandle *)malloc(sizeof(CredHandle));
		if (pcred_handle) {
			pcred_handle->dwLower = cred_handle.dwLower;
			pcred_handle->dwUpper = cred_handle.dwUpper;
		}
	}
	
	*vpcred_handle = pcred_handle;
	return maj_stat;
}

/**********************************************************/

static OM_uint32
gssklog_sspi_init_sec_context_internal(AUTH_METHOD * method,
			OM_uint32 * minor_status,
            void * vpcred_handle,
            void ** vpcontext_handle,
            int  peer_socket,
            char * service_princ_name) 

{
   const char this_function[] = "gssklog_sspi_init_sec_context_internal()";

   OM_uint32 maj_stat = 0;
   gss_buffer_desc input_buffer_desc  = {0,NULL};
   gss_buffer_desc output_buffer_desc = {0,NULL};
   gss_buffer_t output_buffer = &output_buffer_desc ;
   gss_buffer_t input_buffer  = &input_buffer_desc ;
	
   PCredHandle pcred_handle = (PCredHandle)vpcred_handle;
   CtxtHandle newcontext_handle = {0,0};
   PCtxtHandle context_handle = NULL;
   PCtxtHandle xcontext_handle = NULL;
   OM_uint32 ret_flags;
   TimeStamp expiry;
   
   SecBuffer send_tok, recv_tok;
   SecBufferDesc input_desc, output_desc;
   unsigned long deleg_flag;

   int       msg_result ;
   int       rc = -1;
   char * service_princ_name_new = NULL;
   char * cp;

   *minor_status = 0;
   input_desc.cBuffers = 1;
   input_desc.pBuffers = &recv_tok;
   input_desc.ulVersion = SECBUFFER_VERSION;

   recv_tok.BufferType = SECBUFFER_TOKEN;
   recv_tok.cbBuffer = 0;
   recv_tok.pvBuffer = NULL;

   output_desc.cBuffers = 1;
   output_desc.pBuffers = &send_tok;
   output_desc.ulVersion = SECBUFFER_VERSION;

   send_tok.BufferType = SECBUFFER_TOKEN;
   send_tok.cbBuffer = 0;
   send_tok.pvBuffer = NULL;

   deleg_flag =  ISC_REQ_MUTUAL_AUTH |
                 ISC_REQ_ALLOCATE_MEMORY |
                 ISC_REQ_CONFIDENTIALITY |
                 ISC_REQ_REPLAY_DETECT ;

   /* need to convert the service name from service@host to
    * service/host
    */

	service_princ_name_new = strdup(service_princ_name);
	cp= strchr(service_princ_name_new,'@');
    if (cp) *cp = '/';

   /*
    * Perform the context-establishement loop.
    */

   do
   {
      maj_stat =
      InitializeSecurityContext(
                               pcred_handle,
                               context_handle,
                               service_princ_name_new,
                               deleg_flag,
                               0,          // reserved
                               SECURITY_NATIVE_DREP,
                               &input_desc,
                               0,          // reserved
                               &newcontext_handle,
                               &output_desc,
                               &ret_flags,
                               &expiry
                               );
#ifdef DEBUG
    fprintf(stderr,
        "N %s: Returned from  InitializeSecurityContext w/token [%d]\n",
        this_function, send_tok.cbBuffer) ;
#endif /* DEBUG */

      if (recv_tok.pvBuffer)
      {
         free(recv_tok.pvBuffer);
         recv_tok.pvBuffer = NULL;
         recv_tok.cbBuffer = 0;

      }

	  context_handle = &newcontext_handle;

      if (maj_stat!=SEC_E_OK && maj_stat!=SEC_I_CONTINUE_NEEDED)
		goto err;

      if (send_tok.cbBuffer != 0)
      {
		 output_buffer->length = send_tok.cbBuffer;
		 output_buffer->value =  send_tok.pvBuffer;
		 msg_result = send_message(method,
				peer_socket,
				&context_handle,
				output_buffer,
				0, /* 0 = no wrap */
				0) ; /* 0 = no encrypt */

         if (msg_result < 0)
         {
            maj_stat = 3;
			goto err;
         }
#ifdef DEBUG
         fprintf(stderr,
          "N %s: Sent output token [%d], waiting for new token\n",
          this_function, send_tok.cbBuffer) ;
#endif


         FreeContextBuffer(send_tok.pvBuffer);
         send_tok.pvBuffer = NULL;
         send_tok.cbBuffer = 0;
	  }

      if (maj_stat == SEC_I_CONTINUE_NEEDED)
      {
		msg_result = receive_message(method,
			peer_socket,
			&context_handle,
			input_buffer,
			0) ; /* 0 = no unwrap */


         if (msg_result < 0 )
         {
#ifdef DEBUG
			fprintf(stderr,
				"E %s: receive_message: [%d]\n",
				this_function,
				msg_result);
#endif

            maj_stat = 5;
			goto err;
         }

		 recv_tok.cbBuffer = input_buffer->length;
		 recv_tok.pvBuffer = input_buffer->value;

#ifdef DEBUG
      fprintf(stderr,
          "N %s: Received token: [%d]\n",
          this_function,
          recv_tok.cbBuffer) ;
#endif

      }
   } while (maj_stat == SEC_I_CONTINUE_NEEDED);


	
	xcontext_handle = (CtxtHandle *)malloc(sizeof(CtxtHandle));
		if (xcontext_handle) {
			xcontext_handle->dwLower = context_handle->dwLower;
			xcontext_handle->dwUpper = context_handle->dwUpper;
		}
	*vpcontext_handle = xcontext_handle;

	if (service_princ_name_new)
		free(service_princ_name_new);

	return maj_stat;

err:
	if (service_princ_name_new)
		free(service_princ_name_new);

	if (context_handle)
		DeleteSecurityContext(context_handle);

    return maj_stat;
}

/**********************************************************/

static int 
find_realm_of_host_via_dns( char * hostname, char **realm) {

	DNS_STATUS status;
	PDNS_RECORD drhead = NULL;
	PDNS_RECORD pdr;
	char * txt;
	char *hostx;
	int found = 0;
	char hname[DNS_MAX_NAME_BUFFER_LENGTH];

	*realm = NULL;

	for (hostx = hostname; hostx && !found; hostx=strchr(hostx,'.')) {
		hostx++;
		strcpy(hname,"_kerberos.");
		strcat(hname,hostx);

		status = DnsQuery(hname,
					DNS_TYPE_TEXT,
					0,
					NULL,
					&drhead,
					NULL);

		if (status == 0) {
			for (pdr = drhead; pdr && !found; pdr=pdr->pNext) {
				if (pdr->wType == DNS_TYPE_TEXT) {
					txt = pdr->Data.TXT.pStringArray[0];
					*realm = strdup(txt);
					found = 1;
				}
			}
		}
		if (drhead)
			DnsRecordListFree(drhead, DnsFreeRecordList);
	}

	return 1-found;
}

/**********************************************************/

static OM_uint32
gssklog_sspi_init_sec_context(AUTH_METHOD * method,
			OM_uint32 * minor_status,
            void * vpcred_handle,
            void ** vpcontext_handle,
            int  peer_socket,
            char * service_princ_name) 

{
   const char this_function[] = "gssklog_sspi_init_sec_context()";

   OM_uint32 maj_stat = 0;
   char * cp;
   char * hostname;
   char * realm = NULL;
   char * service_princ_name_with_realm = NULL;


   *minor_status = 0;

   maj_stat = gssklog_sspi_init_sec_context_internal(method,
				minor_status,
				vpcred_handle,
				vpcontext_handle,
				peer_socket,
				service_princ_name);

	/* SSPI may not try to get a credential till the init_sec_context */
	/* if so, we dont want to try the additional servers */
	if (maj_stat == 0x8009030c)
		*minor_status = -1;

	/* If SSPI returns with this it means it can't determine the
     * the realm of the service. We are pretty sure there is a server
	 * there, as the application should have established a connection
	 * So we need to take a guess at what is the realm of the 
	 * server. MIT Kerberos would have used the krb5.conf or
	 * the DNS TXT records for _kerberos.<hostname> or derived it. 
	 * We can also use Martin Rex's registry entry for
	 * HKLM\SYSTEM\CurrentControlSet\Control\Lsa\Kerberos\HostToRealm
	 */ 
	if (maj_stat == 0x80090303) {
		/* make sure it is service@hostname and 
		 * not service@hostname@realm 
		 */
		if ((cp = strchr(service_princ_name,'@')) 
				&& (strchr(cp+1,'@') == NULL)) {
			hostname = cp++;

			if (!find_realm_of_host_via_dns(hostname, &realm)) {
				service_princ_name_with_realm = 
					malloc(strlen(service_princ_name) 
					+ strlen(realm) + 2);
				if (service_princ_name_with_realm == NULL) {
					goto err;
				}
				strcpy(service_princ_name_with_realm,service_princ_name);
				strcat(service_princ_name_with_realm,"@");
				strcat(service_princ_name_with_realm,realm);
				
   				maj_stat = gssklog_sspi_init_sec_context_internal(
						method,
						minor_status,
						vpcred_handle,
						vpcontext_handle,
						peer_socket,
						service_princ_name_with_realm);
			}
		}
	}
	
err:
	if (realm)
		free(realm);
	if (service_princ_name_with_realm)
		free(service_princ_name_with_realm);

   return maj_stat;
}

/*****************************************************************/
OM_uint32
gssklog_sspi_delete_sec_context(AUTH_METHOD * method,
	OM_uint32 * minor_status,
	void ** vpcontext_handle) 
{

	PCtxtHandle context_handle = *(PCtxtHandle *)vpcontext_handle;
	*minor_status = 0;
	
	if (context_handle) {
		DeleteSecurityContext(context_handle);
	}
	free(context_handle);
	*vpcontext_handle = NULL;
	return 0;
}

/*****************************************************************/
OM_uint32
gssklog_sspi_wrap(AUTH_METHOD * method,
    OM_uint32 *minor_status,
	PCtxtHandle  * pcontext_handle,
	int conf_req_flag,
	gss_qop_t qop_req,
	 gss_buffer_t input_message_buffer,
	int * conf_state ,
	gss_buffer_t output_message_buffer)
{

   /* SecBuffer in_buf, out_buf; */
   SecBuffer wrap_bufs[3];
   SecBufferDesc in_buf_desc;
   SecPkgContext_Sizes sizes;
   OM_uint32 maj_stat;

   *minor_status = 0;

   maj_stat = QueryContextAttributes(
                                    *pcontext_handle,
                                    SECPKG_ATTR_SIZES,
                                    &sizes
                                    );
   if (maj_stat != SEC_E_OK)
   {
      method->auth_display_status(method,"querying context attributes", maj_stat, 0);
      return (-1);
   }


   //
   // Prepare to encrypt the message
   //

#ifdef DEBUG
   fprintf(stderr,"Block Size is %d, inbuffer = %d\n",
          sizes.cbBlockSize, input_message_buffer->length);
#endif

   in_buf_desc.cBuffers = 3;
   in_buf_desc.pBuffers = wrap_bufs;
   in_buf_desc.ulVersion = SECBUFFER_VERSION;

   wrap_bufs[0].cbBuffer = sizes.cbSecurityTrailer;
   wrap_bufs[0].BufferType = SECBUFFER_TOKEN;
   wrap_bufs[0].pvBuffer = malloc(sizes.cbSecurityTrailer);

   if (wrap_bufs[0].pvBuffer == NULL)
   {
      fprintf(stderr,"Failed to allocate space for security trailer\n");
      return (-1);
   }

   wrap_bufs[1].BufferType = SECBUFFER_DATA;
   wrap_bufs[1].cbBuffer = input_message_buffer->length;
   wrap_bufs[1].pvBuffer = malloc(wrap_bufs[1].cbBuffer);

   if (wrap_bufs[1].pvBuffer == NULL)
   {
      fprintf(stderr,"Couldn't allocate space for wrap message\n");
      return (-1);
   }

   memcpy(
         wrap_bufs[1].pvBuffer,
         input_message_buffer->value,
         input_message_buffer->length
         );

   wrap_bufs[2].BufferType = SECBUFFER_PADDING;
   wrap_bufs[2].cbBuffer = sizes.cbBlockSize;
   wrap_bufs[2].pvBuffer = malloc(wrap_bufs[2].cbBuffer);

   if (wrap_bufs[2].pvBuffer == NULL)
   {
      fprintf(stderr,"Couldn't allocate space for wrap message\n");
      return (-1);
   }



   maj_stat = EncryptMessage(
                            *pcontext_handle,
                            conf_state ? KERB_WRAP_NO_ENCRYPT : 0,
                            &in_buf_desc,
                            0);


   if (maj_stat != SEC_E_OK)
   {
      method->auth_display_status(method,"sealing message", maj_stat, 0);
      return -1;
   }

   //
   // Create the mesage to send to server
   //

   output_message_buffer->length = wrap_bufs[0].cbBuffer 
				 		 		 + wrap_bufs[1].cbBuffer 
						 		 + wrap_bufs[2].cbBuffer;
   output_message_buffer->value = malloc(output_message_buffer->length);

   if (output_message_buffer->value == NULL)
   {
      fprintf(stderr,"Failed to allocate space for wrapepd message\n");
      return (-1);
   }

   memcpy(
         output_message_buffer->value,
         wrap_bufs[0].pvBuffer,
         wrap_bufs[0].cbBuffer
         );
   memcpy(
         (PUCHAR) output_message_buffer->value + (int) wrap_bufs[0].cbBuffer,
         wrap_bufs[1].pvBuffer,
         wrap_bufs[1].cbBuffer
         );
   memcpy(
         (PUCHAR)  output_message_buffer->value + wrap_bufs[0].cbBuffer + wrap_bufs[1].cbBuffer,
         wrap_bufs[2].pvBuffer,
         wrap_bufs[2].cbBuffer
         );


   free(wrap_bufs[0].pvBuffer);
   wrap_bufs[0].pvBuffer = NULL;
   free(wrap_bufs[1].pvBuffer);
   wrap_bufs[1].pvBuffer = NULL;

   return 0;
}

/********************************************************************/

OM_uint32
gssklog_sspi_unwrap(AUTH_METHOD * method,
		OM_uint32 * minor_status,
		PCtxtHandle * pcontext_handle,
		gss_buffer_t input_message_buffer,
		gss_buffer_t output_message_buffer,
		int * conf_state,
		gss_qop_t * qop_state)
{

   SecBuffer wrap_bufs[2];
   SecBufferDesc wrap_buf_desc;
   OM_uint32 maj_stat;
   ULONG qop;

   *minor_status = 0;
   wrap_buf_desc.cBuffers = 2;
   wrap_buf_desc.pBuffers = wrap_bufs;
   wrap_buf_desc.ulVersion = SECBUFFER_VERSION;
   wrap_bufs[0].BufferType = SECBUFFER_STREAM;
   wrap_bufs[0].pvBuffer = input_message_buffer->value;
   wrap_bufs[0].cbBuffer = input_message_buffer->length;
   wrap_bufs[1].BufferType = SECBUFFER_DATA;
   wrap_bufs[1].cbBuffer = 0;
   wrap_bufs[1].pvBuffer = NULL;

   maj_stat = DecryptMessage(
                            *pcontext_handle,
                            &wrap_buf_desc,
                            0,                  // no sequence number
                            &qop
                            );
   if (maj_stat != SEC_E_OK)
   {
      method->auth_display_status(method,"unsealing message", maj_stat, 0);
      return (-1);
   }

   output_message_buffer->length = wrap_bufs[1].cbBuffer;
   output_message_buffer->value  = wrap_bufs[1].pvBuffer;

	if (conf_state)  
		*conf_state = (qop == KERB_WRAP_NO_ENCRYPT)? 0 : 1;
	
   return 0;
}

/****************************************************************/
static
OM_uint32 
gssklog_sspi_display_status(AUTH_METHOD * method,
               char * m, 
               OM_uint32 major_status, 
               OM_uint32 minor_status)
{

	fprintf(stderr,
		"SSPI-error %s: major:%8.8x minor:%8.8x\n",
		m, major_status, minor_status);
	if (major_status) {
		LPVOID lpMsgBuf;
		FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER |
						FORMAT_MESSAGE_FROM_SYSTEM |
						FORMAT_MESSAGE_IGNORE_INSERTS,
					  NULL,
					  major_status,
					  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
					  (LPTSTR) &lpMsgBuf,
					  0,
					  NULL);
		fprintf(stderr,"%s\n",lpMsgBuf);
		LocalFree(lpMsgBuf);
	}

	return 0;
}

/**********************************************************/
static
OM_uint32
gssklog_sspi_release_buffer(AUTH_METHOD * method,
		OM_uint32 * minor_status,
        gss_buffer_t buffer)
{
  OM_uint32 major_status;

  if (buffer->value) {
       free(buffer->value);
  }
  buffer->length = 0;
  *minor_status = 0;
  major_status = 0;

  return major_status;
}
/**********************************************************/
static
OM_uint32
gssklog_sspi_release_cred(AUTH_METHOD * method,
		OM_uint32 *minor_status,
		void ** vpcred_handle) 
{
	OM_uint32 maj_stat = 0;
	PCredHandle pcred_handle = *(PCredHandle *) vpcred_handle;
	*minor_status = 0;
	if (pcred_handle) {
		maj_stat = FreeCredentialsHandle(pcred_handle);
		free(pcred_handle);
		*vpcred_handle = NULL;
	}
	return maj_stat;
}

/*************************************************************/
static AUTH_METHOD gss_auth_data = {
    1,
    gssklog_sspi_acquire_cred,
    gssklog_sspi_init_sec_context,
    gssklog_sspi_wrap,
    gssklog_sspi_unwrap,
    gssklog_sspi_display_status,
    gssklog_sspi_release_buffer,
	gssklog_sspi_release_cred,
	gssklog_sspi_delete_sec_context
    };

/**************************************************************/
AUTH_METHOD * gssklog_sspi_init_method()
{
 return (&gss_auth_data);
}
