/*
 * Pam routine to call gssklog to get an AFS token.
 * The gssklog will be using either GSI or KRB5
 * and will expect to find the proxy or ccache
 * either set in the environment or in the default locations
 * 
 * If a previous PAM routine, like pam_krb5 or the sshd
 * has set the name in the pam_env, then we will pass that on
 * via the env. 
 * 
 * Since hpux does not have the pam_env functions, we will
 * assume the previous ssh or pam_krb5 has set the name
 * of the ccache using either the default of /tmp/krb5cc_<uid>
 * or followed some pattern.  We can pass in the pattern
 * to both pam exists. via ccache=<name>, for example
 *  ccache=FILE:/tmp/krb5cc_p%p_u%u 
 * where %p  and %u are replaced with the PID and UID.
 */

#define DEBLOG(fun, msg) \
	if (debug) \
		syslog(LOG_ERR,"pam_gssklog: %s:%s", fun, msg)

		
#include <stdio.h>
#include <pam_appl.h>
#include <pam_modules.h>
#include <syslog.h>

/* 
 * Even though we don't do anything here, 
 * looks like we still need it, if we want 
 * pam_set_creds to be called 
 */

int
pam_sm_authenticate(pam_handle_t *pamh, int flags, int argc,
	    const char **argv)
{

	int debug = 0;
	int i;
	int ret;
	
	for (i = 0; i < argc; i++) {
		if (strcmp(argv[i], "debug") == 0)
			debug++;
	}
#ifdef DEBUG
	fprintf(stderr,"pam_sm_authenticate flag=%d\n",flags);
#endif
	DEBLOG("pam_sm_authenticate","called");

	/*
	 * we dont really authenticate, we just want to make sure
	 * the pam_set_creds is called later, so we need this 
	 * entry 
	 */
	ret = PAM_CRED_INSUFFICIENT;

	return ret;
}

/********************************************************/ 

int 
pam_sm_setcred(pam_handle_t *pamh, int flags, int argc,
           const char **argv)
{

	int debug = 0;
	int set_pag = 1;
	int i;
	int ret;
	char * username = NULL;
    char ** env;

#ifdef __hpux
	char *p;
	char *q;
	char *cache_name = NULL;
	char cache_name_buf[1024];
	char cache_name_env_buf[2048];
	struct passwd   *pw = NULL;
    char *hpenv_path = "PATH=/krb5/sbin:/usr/bin:/bin";
    char *hpenv[] = {0,0,0};
#endif
	
#ifdef DEBUG
	fprintf(stderr,"pam_sm_setcred flag=%d\n",flags);
#endif

	for (i = 0; i < argc; i++) {
		if (strcmp(argv[i], "debug") == 0) {
			debug++;
		}
#ifdef __hpux
		else if (strncmp(argv[i], "ccache=", 7) == 0) {
			cache_name = &argv[i][7]; /* save for later */
		}
#endif
#ifdef __aix
		else if (strcmp(argv[i], "-a") == 0) {
			apagaix_exec = argv[i+1];
			i++;
		}
#endif
		else if (strcmp(argv[i], "-nopag")  == 0) {
			set_pag = 0;
		}
	}

	DEBLOG("pam_sm_setcred", "called");
	
/* only get new tokens, 
 * TODO: look at restablishing the token, but not the PAG
 * which could be usefull for xlock
 */
	if (!(flags & PAM_ESTABLISH_CRED)) {
		ret = PAM_SUCCESS;
		goto err;
	}

	if (pam_get_item(pamh, PAM_USER, (void **) &username)) {
		ret = PAM_SERVICE_ERR;
		goto err;
	}

#ifdef __hpux

	/* HP does not have the pam_env routines */
	/* will look for ccache as set in krb5 pam routines */

    /* Get username */
	if (pam_get_item(pamh, PAM_USER, (void **) &username)) {
		ret = PAM_SERVICE_ERR;
		goto err;
	}
	if (!(pw = getpwnam(username))) {
		ret = PAM_USER_UNKNOWN;
		goto err;
	}

	/* Get the cache name */
	if (!cache_name) {
		cache_name = "FILE:/tmp/krb5cc_%u";
	}
	strcpy(cache_name_env_buf,"KRB5CCNAME=");
	p = cache_name_env_buf + strlen(cache_name_env_buf);
	q = cache_name;

	/* convert %u and %p */
	while (*q) {
		if (*q == '%') {
			q++;
			if (*q == 'u') {
				sprintf(p, "%ld", pw->pw_uid);
				p += strlen(p);
			} else if (*q == 'p') {
				sprintf(p, "%ld", getpid());
				p += strlen(p);
			} else {
				/* Not a special token */
				*p++ = '%';
				q--;
			}
			q++;
		} else {
			*p++ = *q++;
		}
	}

	hpenv[0] = hpenv_path;
	hpenv[1] = cache_name_env_buf;
	env = hpenv;
#else
	env = pam_getenvlist(pamh);
#endif

    gssklog_pag_klog(set_pag, env);

	
	/* We always return success, even if gssklog failed */
	ret = PAM_SUCCESS;
err:
	return ret;
}

